;;; msdos-shell-fix.el --- MS-DOS COMMAND.COM shell support.

;; Francis Wright <F.J.Wright@QMW.ac.uk>, 22 December 1997

;; Patch the standard comint-mode IF NECESSARY to work with the MS-DOS
;; COMMAND.COM shell via cmdproxy in GNU Emacs 19.34.6 under Windows
;; 95.  It avoids leaving a disconnected shell process if the buffer
;; or emacs is killed, and it arranges for tidy output.  It is
;; intended to work for both command shell buffers and TeX shell
;; buffers (which is the reason for some of the apparent redundancy in
;; the initialization code at the end of this file!)

;; I suggest that you put this library somewhere in your `load-path',
;; compile it and hang it on `comint-exec-hook' by putting this in
;; `.emacs', so that it is loaded only if necessary:

;; (add-hook 'comint-exec-hook
;;           (function (lambda () (require 'msdos-shell-fix))))

;; instead of

;; (setq comint-process-echoes t)
;; (add-hook 'comint-output-filter-functions 'comint-strip-ctrl-m nil t)

;; which it automates.

;; Note that there are two small bugs in the standard 19.34 comint.el
;; that prevent comint-process-echoes and comint-exec-hook (and hence
;; this patch) working properly.

;; If this patch does not work properly then try making the following
;; change to comint.el.  (It has been tested only with versions of
;; comint.el with this change.)

;; Change  (make-local-variable 'comint-exec-hook)
;; to      (make-local-hook 'comint-exec-hook)


(defun msdos-shell-processp (proc)
  "True if PROC is a process running cmdproxy."
  (and (processp proc)
       (let ((case-fold-search t))
	 (string-match "cmdproxy" (car (process-command proc))))
       ))

(defun msdos-shell-send-exit (proc)
  "If PROC is an MS-DOS shell process then send it the string `exit'."
  (if (msdos-shell-processp proc)
      (process-send-string proc "exit\n")))

(defun msdos-shell-send-exit-and-pause (proc)
  "If PROC is an MS-DOS shell process then send it the string `exit'
and pause to allow it to take effect (before Emacs dies)."
  (cond ((msdos-shell-processp proc)
	 (process-send-string proc "exit\n")
	 ;; The time may need tuning on different systems.
	 ;; For me, anything less than 1 second is unreliable.
	 (sit-for 1 nil t)		; no redisplay
	 )))

(defun msdos-shell-kill-emacs-query-function ()
  ;; A function on kill-emacs-query-functions should return a boolean
  ;; value to determine whether to kill emacs.  It may query the user,
  ;; but this version does not!
  (mapcar 'msdos-shell-send-exit-and-pause (process-list))
  t)

(defun msdos-shell-kill-buffer-query-function ()
  ;; A function on kill-buffer-query-functions should return a boolean
  ;; value to determine whether to kill the buffer.  It may query the
  ;; user, but this version does not!
  (msdos-shell-send-exit (get-buffer-process (current-buffer)))
  t)

(defun msdos-shell-bufferp ()
  "True if current buffer has a process running COMMAND.COM."
  (msdos-shell-processp (get-buffer-process (current-buffer))))

(defvar comint-process-echoes)		; for the byte compiler!

(defun msdos-shell-fix ()
  "Install MS-DOS COMMAND.COM shell support if appropriate."
  (if (msdos-shell-bufferp)
      (progn
	(add-hook 'kill-buffer-query-functions
		  'msdos-shell-kill-buffer-query-function)
	(add-hook 'kill-emacs-query-functions
		  'msdos-shell-kill-emacs-query-function)
	(add-hook 'comint-output-filter-functions
		  'shell-strip-ctrl-m nil t)
	(make-local-variable 'comint-process-echoes)
	;; ... done in comint-mode but not in tex-mode.
	(setq comint-process-echoes t)
	)))

(msdos-shell-fix)			; run now and also later!
(add-hook 'comint-exec-hook 'msdos-shell-fix)

(add-hook 'comint-mode-hook 'msdos-shell-fix)
;; ... for shell mode, which declares comint-process-echoes local in
;; comint-mode AFTER running comint-exec

(provide 'msdos-shell-fix)

;;; msdos-shell-fix.el ends here
