;; @(#) dragbar-time.el -- Display incoming mail, etc, in an x-windows drag bar

;; Copyright (C)    1996 TJICorcoran, Jari Aalto
;; Author:	    TJICorcoran   <tjic@icd.teradyne.com>
;; Maintainer:	    Jari Aalto <jari.aalto@poboxes.com>
;; Created:	    May 1994
;; Keywords:	    extensions
;; Compatibility:   Emacs 19.27+, XE 19.14+


;; LCD Archive Entry:
;; dragbar-time|Travis J.I. Corcoran,Jari Aalto|tjic@alta.openmarket.com,jari.aalto@poboxes.com|
;; Displays mail in an x-windows drag bar, Emacs and XE compatible|
;; 05-Jun-1997|$Revision: 1.11 $|~/packages/dragbar-time.el|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;; Commentary:
;;
;;   What this package does:
;;
;;	Allows you to display data periodically in dragbar.
;;	Currently shows you last email arrival and the pending mail count
;;
;;	    "MisterFoo@site.com 6"
;;
;;	See variable db-format-string what other info is displayed.
;;	For non windowed emacs, the information is displayed in the
;;	echo area instead of X dragbar.
;;
;;   How to install:
;;
;;      (require 'dragbar-time)
;;
;;	Or use autoload and you .emacs will be loaded faster (Emacs specific,
;;	XEmacs users must find some similar function). This will start
;;	dragbar after you init fil;e has been loaded.
;;
;;	(autoload 'db-install "dragbar-time" t t)
;;	(run-at-time "2 min" nil 'db-install)
;;
;;   How to use:
;;
;;      Sit back and watch the dragbar tell you incoming mail etc.
;;
;;
;;; Change Log:
;;
;;    5 Jun 97: Jari Aalto       : minor cosmetic changes, minor bugs
;;   14 Nov 96: Jari Aalto       : Reloading now deletes/restarts timer.
;;   13 Nov 96: Jari Aalto       : -nw, No display if minibuffer is active
;;   10 Nov 96: Jari Aalto       : Added autoload installation
;;    6 Nov 96: Jari Aalto       : Now works in Non-windowed/windowed
;;   29 Oct 96: Jari Aalto       : rewrote whole package, made XE compatible
;;   19 May 95: Jari Aalto       : more generic from(1) function
;;   17 May 95: TJICorcoran      : more cleanup
;;   17 May 95: Ake Stenhoff,
;;              Jari Aalto       : multiframe handling
;;   11 May 95: Jari Aalto,
;;              Benoit Archieri,
;;              Ali R. Naddaf    : added email display
;;   30 Apr 95: TJICorcoran      : more cleanup
;;   29 Apr 95: Jari Aalto       : packaged, cleanup
;;          94: TJICorcoran      : wrote basic code
;;
;;   TJICorcoran     == tjic@icd.teradyne.com
;;   Jari Aalto      == jaalto.tre.tele.nokia.fi
;;   Ali R. Naddaf   == naddaf@math.ias.edu
;;   Benoit Archieri == archieri@everest.enst-bretagne.fr
;;   Ake Stenhoff    == etxaksf@aom.ericsson.se
;;

;;; Code:

(require 'tinylibm)

;;; ....................................................... &v-version ...

(defvar db-version-id
  "$Id: dragbar-time.el,v 1.11 1997/06/05 05:37:09 jaalto NotReleased $"
  "RCS version identification string.")

;;; ....................................................... &v-private ...

(defvar db-:old-frame-string     nil "private")
(defvar db-:old-mail-info-string nil "private")

(defvar db-:timer-object nil
  "When package is activated this holds the timer object ativated.")

(defvar db-:mail-info-string nil
  "Mail message information string. This variable has one leading and trailing
space around the message.")

;;; ................................................... &v-user-config ...


(defvar db-:load-hook  'db-install
  "*Hook run when package is loaded.")


;; see from(1), make sure it's on your path, possibly in /usr/ucb/ or
;; /usr/bin/. If you use absolute path, this program executes faster.
;;
(defvar db-:mail-notify-program "from"
  "*Program name that should return entries in the mail spool in
the following format:
   From login@site.xx Mon Feb 26 14:41:50 EET 1996")



(defvar db-:keep-intact-list
  '("VM")
  "A list with frame names not to change.")

(defvar db-:no-mail-string
  (if (ti::xe-window-system)
      " ----"
    ;;  This is better for echo area in non-window emacs
    "-- No Mail --")
  "*String to be printed to dragbar when no mail is pending")

(defconst db-:format-string
  '(concat
    db-:old-frame-string
    " "
    ;;  Use (display-time) in you ~/.emacs to define display-time-string
    ;;
    (if (boundp 'display-time-string)	;may not exist ?
	display-time-string "")
    db-:mail-info-string
    )
  "*Customize your dragbar layout here.")

(defvar db-:mail-info-shorten-regexp nil
  "*When you're in local host and receive mail internally, you
propably want to display user's account name only instead of full
email name. This is REGEXP that is tried upon arrived email address,
if it matches, the email address is truncated to account name.")



;;; ########################################################### &funcs ###

;;; ----------------------------------------------------------------------
;;;
(defun db-get-email-word (str)
  "Returns first word, separated by space."
  (let* ((word str)			;set default
	 )
    (if (null (string-match "From \\([^ ]+\\) " str)) nil
      (setq word (substring str (match-beginning 1)  (match-end 1)))
      )
    word
    ))



;;; ----------------------------------------------------------------------
;;;
(defun db-break-email (str)
  "Break email into two words.
Return:
   (ACCOUNT SITE)  or nil
"
  (let* (w1
	 w2
	 ret
	 )
    (if (null (string-match "[@!]" str)) nil
      (setq w1 (substring str 0 (match-beginning 0)))
      (setq w2 (substring str (1+ (match-beginning 0))))
      ;;   Some sites has "from" command that sends the info in format:
      ;;     "From site.com!login Mon Feb 26 15:50:18 1996"
      ;;
      ;;   And not in traditional format
      ;;     "From login@site.com Mon Feb 26 15:50:18 1996"
      ;;
      ;;   We have to swap the order
      ;;
      (cond
       ((string-match "!" str)
	;; then swap order, since word1 = site, w2 = account
	;;
	(setq ret (list w2 w1)))
       (t
        (setq ret (list w1 w2))))
      )
    ret
    ))



;;; ----------------------------------------------------------------------
;;;
(defun db-mail-info ()
  "Runs proses from(1) to get informtion about arrived mail.

Returns:
   list       (line ..)

where LINEs are all full from(1) entries received: oldest first.
If no mail in spool, returns nil.
"
  (let* ((default-directory  default-directory)
	 (buffer	     (generate-new-buffer "*mail-spool*"))
	 (tmp-dir	     "/tmp/")
	 (mail-proc	     db-:mail-notify-program)
	 ret line
	 )
    (save-excursion
      (set-buffer buffer)
      (unwind-protect
	  (progn
	    ;; - launch up the process and restore the directory setting
	    ;; - The output is like:
	    ;;
	    ;;   From naddaf@math.ias.edu Thu May 11 18:55:59 EET 1995
	    ;;   From akontiai@tre.tele.nokia.fi Thu May 11 19:05:36 EET 1995
	    ;;
	    (setq default-directory tmp-dir)
	    (call-process mail-proc nil buffer nil)     ;;wait till finishes...

	    (unless (eq (point-min) (point-max))	;No output ?
	      ;;  - Now read persons email delimited by spaces
	      ;;  - Read the last line to get newest mail arrival
	      ;;
	      (goto-char (point-min))
	      (while (not (eobp))
		(setq line
		      (buffer-substring
		       (point)
		       (progn (end-of-line) (point))))
		(setq ret (append ret (list line)))
		(forward-line 1)
		)))
	(if (get-buffer buffer)		;Buffer exist ?
	    (kill-buffer buffer))
	)
      ;; ....................... unwind end
      ret
      )))


;;; ----------------------------------------------------------------------
;;;
(defun db-get-mail-info-string ()
  "Returns mail string: last sender and mail count"
  (let* ((list		(db-mail-info))
	 (re		db-:mail-info-shorten-regexp)
	 (ret		db-:no-mail-string)	;default return
	 last-line
	 email count
	 )
    (cond
     (list				;only if mail is waiting
      (setq count	(length list)
	    last-line	(car (reverse list))
            email	(db-get-email-word last-line)
	    )

      ;;  now see if user wants shortened version ?
      (cond
       ((and (stringp re)			;no regexp
	     (string-match re email)
	     (setq list (db-break-email email)))
	(setq email (nth 0 list))))

      (setq ret (concat " " email " " count))
      ))
    ret
    ))


;;; ----------------------------------------------------------------------
;;;
(defun db-update ()
  "Updates the frame's status line, or in non-X show the message
in echo area."
  ;;   only update the bar when needed,
  ;;   no matter how often this function
  ;;   is called
  (let* ((str "")
	 flist
	 elt
	 )
    ;;  Save the contents of frame name, eg. host name only once
    ;;
    (if (null db-:old-frame-string)
      (setq db-:old-frame-string
	    (ti::xe-set-frame-name nil nil 'get)))

    (setq db-:mail-info-string
	  (format " %s " (db-get-mail-info-string)))


    (cond
     ((ti::xe-window-system)
      (setq flist (frame-list))
      ;; Update frames that are not exluded
      ;;
      (list-loop flist elt
	(if (not (member
		  (ti::xe-set-frame-name nil nil 'get)
		  db-:keep-intact-list))
	    (ti::xe-set-frame-name (eval db-:format-string) elt)
	  )))
     (t
      ;; Non-windowed emacs
      ;;
      (if (not (string= db-:mail-info-string db-:no-mail-string))
	  (setq str "Mail notification: "))

      (cond
       ((and (not (ti::xe-executing-macro))
	     ;; Having this mode operate in the minibuffer
	     ;; makes it impossible to see what you're doing.
	     (not (eq (selected-window) (minibuffer-window)))
	     (sit-for 0.50)
	     )
	(message "%s%s" str (eval db-:format-string))
	(sleep-for 1)			;make sure user sees it
	))
      ))


    ;; Notify about new mail ?
    ;;
    (cond
     ((and (stringp db-:old-mail-info-string)
	     (not (string= db-:mail-info-string db-:old-mail-info-string)))
      (beep)
      (sit-for 0.3)
      (beep)
      ))

    (setq db-:old-mail-info-string db-:mail-info-string)
    ))

;;; ----------------------------------------------------------------------
;;;
(defun db-install ()
  "Install dragbar Mail watchdog. The installation chosen depends on Emacs
version."
  (interactive)
  ;; In XEmacs the frame must be configured by hand
  ;;
  (when (and (ti::xe-window-system)
	     (xemacs-p))
    ;;  make sure it's list
    ;;
    (setq frame-title-format (list-make frame-title-format))
    (unless (memq 'db-:mail-info-string frame-title-format)
      (push 'db-:mail-info-string frame-title-format))
    )


  (cond
   ((fboundp 'run-with-idle-timer)	;19.34+

    ;;   Every 20th second we show the mail status
    ;;   Delete old process.
    (if db-:timer-object		;Start only once
	(safe-do (sfuncall 'cancel-timer db-:timer-object)))

    (setq db-:timer-object
	  (sfuncall 'run-with-idle-timer 20 'repeat 'db-update))
    )

   (t
    ;; This is Emacs call, but the 'xe' lib can pass it to XE too
    ;; Let's not run so often...
    ;;
    (if db-:timer-object		;Start only once
	(ti::xe-timer-key-delete db-:timer-object))


    (setq db-:timer-object (run-at-time "1 min" 60 'db-update))
    ))

  ;;  function end
  )


(provide   'dragbar-time)
(run-hooks 'db-:load-hook)

;;; dragbar-time.el ends here
