;; @(#) tinylock.el -- Simple emacs locking utility

;; @(#) $Id: tinylock.el,v 1.28 1997/06/05 08:13:23 jaalto NotReleased $
;; @(#) $Keywords: extensions $
;;
;; $KnownCompatibility: 19.28+, XEmacs 19.14+ $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs

;;{{{ Id

;; Copyright (C) 1995,96,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Feb 11 1995
;;
;; To get information on this program use ident(1) or do M-x til-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el

;; LCD Archive Entry:
;; tinylock|Jari Aalto|jari.aalto@poboxes.com|
;; Simple emacs locking utility|
;; 05-Jun-1997|$Revision: 1.28 $|~/misc/tinylock.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;; ....................................................... &t-install ...
;; Put this file on your Emacs-Lisp load path, add one of these into your
;; ~/.emacs startup file
;;
;;   Normal load
;;
;;	(require 'tinylock)
;;
;;   Autoload, your emacs starts up faster, prefered, but doesn't
;;   activate the autolocking feature.
;;
;;	(autoload 'til-lock "tinylock" "Lock emacs" t)
;;
;;   ESC-l, suggested keybinding, replaces downcase-word binding
;;   because you can accomplish the same with C-x C-l,
;;   downcase-region.
;;
;;	(global-set-key "\M-l" 'til-lock)     ;; Suggested keybinding.
;;
;;
;;   If you have any questions, use this function
;;
;;      M-x til-submit-feedback
;;
;;   See also Example section at the end of file.

;;}}}
;;{{{ Documentation

;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVERD prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "til-" in front of
;;   every function & variable. It stands for '(ti)ny (l)ock.
;; - variable names contain letters '-:', except version/hook vars.

;;  Preface, feb 1995
;;
;;      This is again one of those 'tiny tools' I came up with time to
;;      time. It was inspired by a post to gnu.emacs.help group where
;;      someone asked for emacs locking capability. 2 years ago I did
;;      similar terminal locking shell script, but I never thought it
;;      ewould be nice for emacs too.
;;
;;      I had a sketch of what to do 2 months ago, but I was so busy I
;;      never completed this el. So I couldn't post the answer to
;;      person asking for the locking capability.
;;
;;      I find this usefull, when I'm off work and want to set my
;;      emacs in consistent state, so that no-one accidentally messes
;;      my C++ project files. I normally leave all xterms and emacses
;;      open during the weekend and anyone may use my X-tube
;;      meanwhile. And nowadays when I use PGP I appreciate that I can
;;      lock my mail buffers and PGP pass phrases inside emacs safely when
;;      I go somewhere else for a while. No-one can snoop my emacs
;;      while I'm away.
;;
;;      Hope people find this usefull!
;;
;; Overview of features
;;
;;	o   Locks emacs completely until right key is entered.
;;	o   Auto-locks emacs after XXX idle time
;;	o   False attemps are stored in history log.
;;	o   Blanks or displays buffer message when locked.
;;	o   Hooks: before and after lock is entered/removed
;;
;;  About locking procedure
;;
;;      Don't get shocked now... When the lock gets in effect there must
;;      be no running processes inside emacs that would generate
;;      error and make emacs vulnerable to break in's.
;;
;;      That's why all the running processes are killed before the
;;      lock takes in effect. If you have some valuable
;;      processes that are constantly running, you must make a
;;      separate "process control" function that would restart any
;;      such processes. Use the appropriate hook to activate those
;;      processes again after the emacs is unlocked. Use hooks
;;
;;	    til-:before-lock-hook       ;; Save processes here
;;	    til-:after-lock-hook        ;; restore processes here
;;
;;      and function
;;
;;	    M-x list-processes
;;
;;      which tells you what processes you're running.
;;
;;      All extra frames are also deleted. At least for now, because
;;      I don't know a reasonable way to save the frame configurations
;;      at this time. Maybe later when I have more time I investigate
;;      how do I blank the other frames and restore them after unlock.
;;
;;      Please send me piece of code if you know good solution.
;;
;;  About autolocking feature, Emacs prior 19.34
;;
;;      When you load this package the til-:load-hook runs
;;	til-process-control command that setup up a timer process
;;	that wakes up periodically. If the emacs has not changed
;;	compared to last saved emacs state, then the autolocking
;;	takes in effect immediately.
;;
;;      Currently the emacs activity is determined in simple way
;;
;;      o if buffer list order has changed user is doing something.
;;      o if switch-buffer is used, user is doing something
;;      o if any buffer's size is changed, user is doing something.
;;
;;      This checking may not be anough in somes: if user
;;      just scroll some text in buffer for XXX
;;      minutes, then from til-process's point of view there has
;;      not been any activity and the user may suddenly
;;      notice that emacs locks up.
;;
;;      Doing nothing but viewing one buffer all the time is
;;      fortunately rare. If you have more ideas how to detect
;;      emacs activity better, mail me more suggestions.
;;
;;  About autolocking feature, 19.34+
;;
;;	New Emacs releases have command run-with-idle-timer which
;;	we use if it is available. When there has been no activity
;;	for XXX minutes, your Emacs locks up.
;;
;;	The advice code and the other tricks we needed to detect idle
;;	activity in lower emacs versions aren't installed or needed
;;	in these Emacs versions, so you don't have to worry about them.
;;
;;  Autolock password
;;
;;      Do not put password in your ~/.emacs, but answer to the
;;	question which is asked when this file is loaded. If you
;;	want to change it during your emacs session, call function
;;
;;	    M-x til-auto-lock-set-password
;;
;;  Changing the autolock interval
;;
;;	The autolock interval depends on the wake up time of timer
;;	process. The default time is 20 minutes when you load this
;;	file. You can change the time by calling
;;
;;	    M-x til-auto-lock-set-interval
;;
;;     Or by putting this code in your ~/.emacs
;;
;;* _
;;*        ;; First define the hook, so that we can append to it
;;*        (setq til-:load-hook '(til-timer-control til-auto-lock-set-password)
;;*
;;*        ;; add function to the end
;;*        (add-hook 'til-:load-hook 'my-til-auto-lock-set-interval 'append)
;;*
;;*
;;*        (defun my-til-auto-lock-set-interval ()
;;*          "Change interval to 10 minutes."
;;*          (til-auto-lock-set-interval 10))

;;}}}
;;{{{ History

;; ........................................................ &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; May 31	1997	[jari]		19.28	v1.27-1.28	NotReleased
;; - lib 'm' function changes. Checkdoc 1.42 clear. defcustom support.
;; - Library function name change.
;;
;; May 16	1997	[jari]		19.28	v1.24-1.26	NotReleased
;; - Internal changes. Compilation changes.
;; - Changed all non standard boolean tests to use when and unless
;;
;; Mar 21	1997	[jari]		19.28	v1.21-1.23	NotReleased
;; - Couldn't lock the window if I was in a dedicated frame.
;;   Now corrected.
;;
;; Jan 15	1997	[jari]		19.28	v1.18-1.20	NotReleased
;; - Updated code to use 19.34's idle timers. Ther timer handling was
;;   changed in 19.34 --> updated calls to tinylib.el
;;
;; Nov 9	1996	[jari]		19.28	v1.15-1.17	NotReleased
;; - One library function name changed. autoload are in 'm' lib now.
;; - Added error protection to til-user-activity, so that if data
;;   is corrupted, then it initalizes it.
;;
;; Sep 27	1996	[jari]		19.28	v1.12-1.14	NotReleased
;; - Rewrote most of the code. Removed unnecessary configuration variables.
;; - Added autolocking feature.
;;
;; Sep 25	1996	[jari]		19.28	v1.10-1.11	NotReleased
;; - Rearranged loading. Now uses autoloads.
;; - Made this package work in XEmacs too.
;;
;; Aug 30	1996	[jari]		19.28	v1.9		NotReleased
;; - Cleared all XEmacs 19.14 ByteComp errors. This file doesn't work in
;;   XEmacs until I figure out read-event command in XEmacs.
;;
;; Jun 20	1996	[jari]		19.28	v1.8		NotReleased
;; - Added safe-do when runnig processes down...
;; - Frame handling missing. Now it kills extra frames. Sorry, your
;;   your own frame config saver in some hook if tou want to save
;;   and resore them when unlocked.
;;
;; Feb 10	1996	[jari]		19.28	v1.7		NotReleased
;; - lisp-mnt.el's lm-verify checked. ByteComp ok. Rewrote feedback funcs.
;;
;; Nov 17	1995	[jari]		19.28	v1.6		NotReleased
;; - Started preparing functions to use my tinylibXX.el libraries to avoid
;;   function duplication. Well, code shrunk remarkably :-)
;;
;; Sep 1	1995	[jari]		19.28	v1.5		Release_4
;; - If there were running process that could call 'error', like GNUS,
;;   this program would break and another user could access the emacs.
;;   Now we kill all running processes, before shutting down to prevent any
;;   other tasks to interrupt us.
;; - Added ###autoload directives. Added til-:load-hook.
;;
;; Mar 7	1995	[jari]		19.28	v1.4		Release_3
;; - Added til-:blank-kill, so that it wont't crowd buffer-menu.
;; - Added til-:h-limit + several funcs to limit locking time if needed.
;;
;; Mar 1	1995	[jari]		19.28	v1.3		NotReleased
;; - Added 'blank' capability and restoring the screen. Throw err. fixed.
;; - autoload setup presented. Added til-[before,after]-lock-hook
;;
;; Feb	13	1995	[jari]		19.28	v1.2		Release_2
;; - til-history: Now keeps record of false enries,
;; - password input is now more bullet proof, has new parameter to limit input.
;; - Uses events for reading confirm now, now you can use mouse also :-/
;; - Corrected the library load, using unwind-protect now. Prg Date corr.
;; - Fixed one TIR- function to use modulename TIL-. Wait time corrected.
;;
;; Feb	11	1995	[jari]		19.28	v1.1		Release_1
;; - First *complete* implementation.

;;}}}

;;; Code:

;;{{{ setup: require

;;; ......................................................... &require ...

(require 'tinylibm)


(eval-and-compile
  ;; If this is not 19.34+, then we need advice code, otherwise it is
  ;; skipped.
  ;;
  (unless (fboundp 'run-with-idle-timer)  (require 'advice))
  (use-timer)
  )

(eval-when-compile (use-dynamic-compilation))


(defgroup-tiny TinyLock til-: extensions
    "Simple emacs locking utility.
 Overview of features

	o   Locks emacs completely until right key is entered.
	o   Auto-locks emacs after XXX idle time
	o   False attemps are stored in history log.
	o   Blanks or displays buffer message when locked.
	o   Hooks: before and after lock is entered/removed"
    )



;;}}}
;;{{{ setup: variables

;;; ......................................................... &v-hooks ...


(defcustom til-:load-hook nil
  "*Hook run after file is loaded."
  :type 'hook :group 'TinyLock
  )

(defcustom til-:before-lock-hook  nil
  "*Hook that is run when the locking process initiates.
This is your chance to save frame configurations or processes before
they are killed."
  :type 'hook :group 'TinyLock
  )


(defcustom til-:after-lock-hook nil
  "*Hook that is run after lock is removed."
  :type 'hook :group 'TinyLock
  )

;;; ....................................................... &v-private ...

(defconst til-:history nil
  "\(DATE PASSWD\) A storage where attempts of entering locked emacs is put.
Cleared every time lock takes effect.")


(defvar til-:auto-lock-data  nil
  "Data to tell about the emacs idle state, updated by timer process.
Contains:
'(current-time          ;; time stamp of user activity
  (BUFFER-LIST)
  (SIZE SIZE SIZE ..)   ;; every buffers size.
  )")


(defvar til-:auto-lock-password  nil
  "Password in auto lock situation.
Password is asked when you load this file. You shouldn't define
this in you ~/.emacs")

(defvar til-:auto-lock-interval  nil
  "The timer interval in minutes. Use \\[til-auto-lock-set-interval].")

(defvar til-:idle-timer-process  nil
  "19.34+ timer process.")

;;; ........................................................ &v-config ...

(defcustom til-:entry-err-sleep 2
  "*Time in seconds that is waited until new login to emacs is possible."
  :type '(integer :tag "Seconds") :group 'TinyLock
  )

(defcustom til-:hist-buf "*til-hist*"
  "*Buffer to output the history data."
  :type 'string :group 'TinyLock
  )

(defcustom til-:blank-buiffer "*blank*"
  "*Buffer name used when screen is blanked."
  :type 'string :group 'TinyLock
  )

(defcustom til-:blank-when-locked t
  "*If non-nil, the buffer `til-:blank' is shown, created, if not exist."
  :type 'string :group 'TinyLock
  )


;;}}}
;;{{{ setup: version

;;; ....................................................... &v-version ...

(eval-and-compile
(ti::m-fmacro-version-bug-report
 "tinylock.el"
 "til"
 til-:version-id
 "$Id: tinylock.el,v 1.28 1997/06/05 08:13:23 jaalto NotReleased $"
 '(til-:version-id
   til-:before-lock-hook
   til-:after-lock-hook
   til-:load-hook
   til-:auto-lock-data
   til-:auto-lock-password
   til-:auto-lock-interval
   til-:idle-timer-process
   til-:entry-err-sleep
   til-:hist-buf
   til-:blank-buiffer
   til-:blank-when-locked
   )))


;;}}}


;;{{{ code: macros, advices

;;; ########################################################## &Macros ###

(defmacro til-time-dd (time)
  "Return Day from TIME."
  (list 'string-to-int (list 'substring time 8 10)))

(defmacro til-time-hh (time)
  "Return hour from TIME."
  (list  'string-to-int (list 'substring time -13 -11)))


;;; ######################################################### &advices ###

(eval-and-compile
(unless (fboundp 'run-with-idle-timer)		;we need this if not 19.34+
  (require 'advice)

  ;;   What else easy means we have to tell that user is working with
  ;;   the emacs ?
  ;;
  ;;   The advice shouldn't disturb normal emacs behavior and the functions
  ;;   calls are _inlined_, ie. function is expanded to point
  ;;   when byte compiled, so that the advice works as fast as possible
  ;;   and doesn't take time from the original function.

  ;; (ti::m-ad-control '(switch-to-buffer other-window)  "^til$" 'dis)
  ;;

  (defadvice switch-to-buffer  (before til act) ;C-x C-b
    "Tell to Emacs autolock that there is user activity."
    (if (interactive-p) (inline (til-user-activity))))


  (defadvice execute-extended-command  (before til act)   ;; M-x called
    "Tell to Emacs autolock that there is user activity."
    (if (interactive-p) (inline (til-user-activity))))

  (defadvice other-window  (before til act) ;C-x o
    "Tell to Emacs autolock that there is user activity."
    (if (interactive-p) (inline (til-user-activity))))
  ))

;;}}}
;;{{{ code: misc funcs

;;; ########################################################### &Funcs ###


;;; ----------------------------------------------------------------------
;;;
(defun til-auto-lock-set-interval (minutes)
  "Set new MINUTES interval by stopping and restarting timer process."
  (interactive "Nminutes: ")
  (til-process-control nil minutes)
  nil
  )


;;; ----------------------------------------------------------------------
;;;
(defun til-auto-lock-set-password ()
  "Set auto lock password."
  (interactive)
  (let* (pass
         )
    (if (var-nil-p (setq pass (ti::q-read-input-as-password
			       "TinyLock, give autolock password: ")))
	(error "Password is empty.")
      (setq til-:auto-lock-password pass))
    nil
    ))



;;; ----------------------------------------------------------------------
;;;
(defun til-process-on ()
  "Start auto lock process."
  (til-process-control nil til-:auto-lock-interval))

;;; ----------------------------------------------------------------------
;;;
(defun til-process-control (&optional delete interval)
  "Install process that lock Emacs when there is no activity.
Optionally DELETE til-process. INTERVAL is in minutes, defaults to 20."
  (interactive "P")
  (let* ((elt	    (ti::xe-timer-elt 'til-process))
	 (interval  (or interval 20))	;Default minutes
	 )
    (cond
     ((fboundp 'run-with-idle-timer)	;19.34+
      (cond
       (delete
	;; Not implemented
	(if til-:idle-timer-process
	    (sfuncall 'cancel-timer til-:idle-timer-process))
	)
       (t
	;;  Remove old
	(if til-:idle-timer-process
	    (sfuncall 'cancel-timer til-:idle-timer-process))

	(setq
	 til-:idle-timer-process
	 (sfuncall
	  'run-with-idle-timers (* interval 60) 'repeat 'til-lock-now)
	 ))
       ))
     (t
      (if delete
	  (progn
	    (ti::xe-timer-key-delete elt)
	    (if (interactive-p)
		(message "Autolock process deleted."))
	    )
	(ti::xe-timer-key-delete elt)	;Remove old
	(til-process-data-set)
	(setq til-:auto-lock-interval interval)
	(run-at-time "1 sec" (* interval 60) 'til-process)

	(if (interactive-p)
	    (message "Autolock process activated."))
	)))
    elt
    ))


;;; ----------------------------------------------------------------------
;;;
(defun til-user-activity ()
  "Tell to timer process that the has bee user activity."
  (or
   (safe-do (setcar til-:auto-lock-data (current-time)))
   ;;  Hmm, data is corrupted... reset it.
   (til-process-data-set)
   ))


;;; ----------------------------------------------------------------------
;;;
(defun til-process-data-set ()
  "Update timer process data."
  (setq til-:auto-lock-data
	(list
	 (current-time)
	 (buffer-list)
	 (mapcar
	  (function
	   (lambda (x)
	     (job-buffer x
	       (buffer-size))))
	  (buffer-list)
	  )))
  nil
  )

;;; ----------------------------------------------------------------------
;;; Just testing... (til-process-data-set) (til-process-data-unchanged-p)
;;;
(defun til-process-data-unchanged-p ()
  "Return t if timer data has not changed = No activity in emacs."
  (let* ((data		til-:auto-lock-data)
	 (time          (nth 0 data))
	 (buffer-list	(nth 1 data))
	 (size-list	(nth 2 data))
	 (list		(buffer-list))
	 (i		0)
	 elt
	 unchanged
         )
    (if (null til-:auto-lock-data)
	(til-process-data-set)

      ;; o if buffer list order is the same, the user may not have
      ;;   done any new work...
      ;; o Next we check if buffer sizes have changed, if not, then
      ;;   user hasn't done any work in emacs.
      ;;
      (condition-case nil
	  (and (> (ti::d-time-difference (current-time) time)
		  (- (* til-:auto-lock-interval 60) 5)) ;5 sec timeframe
	       (equal list buffer-list)
	       (progn
		 (setq unchanged t)
		 (list-loop list elt
		   (job-buffer elt
		 (if (not (eq (buffer-size)
			      (nth i size-list)))
		     ;; Found changed buffer, stop there and
		     ;; reset lock status, and quit the loop by
		     ;; killing the list
		     ;;
		     (setq list nil  unchanged nil))
		 (inc i)
		 ))))
	;;  Data is corrupted somehow, fix it.
	(error
	 (til-process-data-set)
	 ))
      unchanged
      )))


;;; ----------------------------------------------------------------------
;;;
(defun til-process ()
  "Locks the emacs if it there has been no user activity."
  (when (til-process-data-unchanged-p)
    ;;  When Emacs locks up, this function process will die too.
    (til-lock-now))
  (til-process-data-set)		;Update the data
  )



;;; ----------------------------------------------------------------------
;;;
(defun til-add-history (passwd)
  "Add to record log `til-:history', any attempts to enter locked emacs.
PASSWD is the attempted login password."
  (let* ((d (current-time-string))
	 )
    (setq til-:history
	  (append  til-:history
		   (list (list d passwd))))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun til-kill-process-control (&optional kill)
  "Return all processes in string format, or KILL all processes (not timer)."
  (let* ((list  (process-list))
	 ret
	 )
    (if list
	(mapcar
	 (function
	  (lambda (x)
	    (cond
	     ((null kill)
	      (setq ret (concat (or ret "") (prin1-to-string x))))
	     (t
	      ;;  let's not kill the timer
	      (if (not (string-match "display-time\\|timer"
				     (prin1-to-string x)))
		  (delete-process x))
	      ))
	    ))
	 list))
    ret
    ))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun til-history (&optional buffer)
  "Displays login history. Optionally to given buffer BUFFER."
  (interactive)
  (let* ((list  til-:history)
	 (i	0)
	 elt
	 d p
	 )
    (if buffer
	nil				;user gave it
      (setq buffer (get-buffer-create til-:hist-buf))
      (set-buffer buffer)			;for insert
      (erase-buffer)
      (switch-to-buffer-other-window buffer))

    (list-loop list elt
      (setq d (nth 0 elt))    (setq p (or (nth 1 elt) "<nil>"))
      (insert (format "%2d: %-27s %s\n" i d p))
      (setq i (1+ i))
      )
    ))



;;; ----------------------------------------------------------------------
;;;
(defun til-blank-control (&optional unblank)
  "Blank display or UNBLANK."
  (let* (frame-list
	 (blank-bp       (get-buffer-create til-:blank-buiffer))
	 elt
         )

    (cond
     (unblank
      (kill-buffer-safe blank-bp))

     (t
      (use-non-dedicated-frame)
      (setq frame-list (delq (current-frame) (frame-list)))

      ;;  Kill other frames
      (list-loop frame-list elt  (delete-frame elt))

      (switch-to-buffer blank-bp t)
      (delete-other-windows)	;delete all other windows
      ;;   This is necessary in 19.28 for some unknown reason
      ;;   otw, the sreen is not shown
      (sit-for 1)
      ))))

;;}}}
;;{{{ code: main


;;; ----------------------------------------------------------------------
;;;
(defun til-lock-now ()
  "Locks emacs."
  (til-lock til-:auto-lock-password "Autolocking.. emacs " 'doit ))


;;; ------------------------------------------------------------ &main ---
;;;
;;;###autoload
(defun til-lock (psw &optional msg lock-now)
  "Lock Emacs with PSW password and MSG.
Message is displayed if LOCK-NOW is nil.
If LOCK-NOW is non-nil emacs is immediately locked with PSW."
  (interactive
   (list
    (progn
      (message-quick "Now enter lock string...")
      (ti::q-read-input-invisible))))

  (let* (
	 (cursor-in-echo-area nil)

	 ;;  It's good programming style NOT to use globals directly
	 ;;  inside code This way maintainer sees at glance what it uses.
	 (key-msg	 "This emacs is locked, enter key:")
	 (entry-err	 "Unauthorized access.")
	 (wait		 til-:entry-err-sleep)
	 (loop		 t)
	 (msg            (or msg "Lock Emacs ? "))
	 ans
	 )
    (catch 'done
      (if (var-nil-p psw)
	  (error "Password is empty."))

      (if (and (null lock-now)
	       (null (y-or-n-p msg)))
	  (throw 'done t))

      (save-window-excursion
	(run-hooks 'til-:before-lock-hook))

      ;;  It's better to save work, you may forgot the password :-/
      (save-some-buffers 'noAsk)

      (ti::xe-timer-list-control    'save)
      (til-process-control	    'kill)	;our process
      (til-kill-process-control	    'kill)	;get rid of them
      (ti::xe-timer-list-control    'kill)	;Stop all timers


      (til-blank-control)
      ;;   we need to restore windows config when we return

      (save-window-excursion
	(save-excursion
	  ;; Now we make interrupting impossible, C-g won't work now on...
	  (setq inhibit-quit t)
	  (setq til-:history nil)	;clear the log buffer

	  (message-quick ">> LOCKED <<")
	  (message "")

	  (while loop
	    (when (input-pending-p) 	;wait for kbd event
	      (discard-input)
	      (message key-msg) (sleep-for 1) (message "")


	      (discard-input)
	      (setq ans (ti::q-read-input-invisible))
	      (cond
	       ((string-equal ans psw)
		(setq loop  nil))	; right password, let user in
	       (t
		(til-add-history ans)	; record to log
		(message entry-err)
		(sit-for wait)))
	      )
	    )))

      (til-blank-control 'unblank)
      (message "Emacs unlocked")
      (setq quit-flag nil inhibit-quit nil)       ; restore flags

      (ti::xe-timer-list-control 'restore)

      (til-process-on)
      (run-hooks 'til-:after-lock-hook)

      nil
      )))

;;}}}
;;{{{ txt: example

;;; ......................................................... &example ...
;;; - Code can be extracted with function tinylib.el/ti::m-pkg-rip-magic
;;;
;;; - Here I enable the time process again and before locking
;;;   I want to save my emacs state. [tinnydesk.el]

;;* _
;;* _
;;* (add-hook 'til-:before-lock-hook 'my-til-:before-lock-hook)
;;* (add-hook 'til-:after-lock-hook  'my-til-:after-lock-hook)
;;* _
;;* _
;;* (defun my-til-:before-lock-hook ()
;;*   "Saves emacs state, so that you can recover from
;;* accidental crash.
;;* "
;;*   (require 'tinydesk)
;;*   (message "Wait, saving state...")
;;*   (tid-save-state "~/wtxt/state.lock")
;;*   (message "Wait, saving state...done")
;;*   )
;;* _
;;* (defun my-til-:after-lock-hook ()
;;*   "Restores some things after lock"
;;*   (display-time)			;re-enable process
;;*   (load "dragbar-time" 'noerr)      ;My mail notification program
;;*   )
;;* _
;;* (autoload 'til-lock	       "tinylock"	t t)
;;* _

;;}}}


(add-hook 'til-:load-hook	'til-process-on)
(add-hook 'til-:load-hook	'til-auto-lock-set-password)

(provide   'tinylock)
(run-hooks 'til-:load-hook)

;;; tinylock.el ends here
