;; @(#) tinypage.el -- Handling ^L pages, select, cut, copy, head renumber...

;; @(#) $Id: tinypage.el,v 1.24 1997/05/29 18:00:03 jaalto NotReleased $
;; @(#) $Keywords: tools $
;; $KnownCompatibility: 19.28, XEmacs 19.14 $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs

;;{{{ Id

;; Copyright (C) 1996,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Jun 10 1996
;;
;; To get information on this program use ident(1) or do M-x txx-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el

;; LCD Archive Entry:
;; tinypage.el|Jari Aalto|jari.aalto@poboxes.com|
;; ^L page select, cut, copy. Renumbering 1.1 etc. levels. No narrowing used |
;; 29-May-1997|$Revision: 1.24 $|~/misc/tinypage.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;; ....................................................... &t-install ...
;;  Put this file on your Emacs-Lisp load path, add following into your
;;  ~/.emacs startup file
;;
;;	(require 'tinypage)
;;
;;  or use this; your .emacs loads quicker. Preferred.
;;
;;	(global-set-key "\C-cmp" 'tip-mode)   ;; "m" for minor modes
;;	(autoload 'tip-mode "tinypage" t t)
;;
;;  If you make any changes to keybindings, always run command
;;
;;	M-x tip-mode-install
;;
;;  to see what this mode offers, look at the mode description
;;
;;	M-x tip-mode
;;	C-h m
;;
;;  If you have any questions, use this function
;;
;;      M-x tip-submit-bug-report       ,send bug report
;;


;;}}}

;;{{{ Documentation

;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "tip-" in front of
;;   every function & variable. It stands for '(ti)ny (P)age handling'.
;; - variable names contain letters 'tip-:'

;;  Preface, jun 1996
;;
;;	I had found paged.el by Michelangelo Grigni <mic@mathcs.emory.edu>
;;	one year or so ago and had liked it very much. Unfortunately
;;	it used narrowing and didn't offer easy page select, copy, cut
;;	actions which belong to basic page editing.
;;
;;	Paged.el has one nice feature: It can renumber pages and make summary
;;	out of them. If I have time I will include those features to
;;	package too.
;;
;;  Overview of features
;;
;;	o   Copy, cut, paste, yank (after/before current page) ^L pages.
;;	o   Show  page-nbr/page-count/page-size in modeline.
;;	o   Can renumber numbered header levels, where last level is indicated
;;	    with number. Eg. "A.1 A.2"  or "1.2.1.1 "1.2.1.2"
;;	o   Shows popup in X to jump to headings
;;	o   Create table of contents.
;;
;;  About making pages -- basics
;;
;;	If you're totally unfamiliar to the concept of page: you make
;;	pages in emacs by adding the linefeed marker in the text, normally
;;	on its own line, just before your topics or headings.
;;
;;	    C-q C-l  --> ^L
;;
;;	That inserts the ^L character in the buffer. That is where your
;;	page starts. The layout of your doc may look like this:
;;
;;         ^L
;;         1.0 Topic one
;;             txt txt txt txt txt txt txt txt
;;             txt txt txt txt txt txt txt txt
;;
;;         ^L
;;         1.1
;;      	txt txt txt txt txt txt ..
;;
;;         ^L
;;         1.1.1.1
;;      	txt txt txt txt txt txt txt txt
;;       	txt txt txt txt txt txt txt ...
;;
;;
;;  About renumbering
;;
;;	This package offers simple renumbering features, but it
;;	won't do everything for you! Let's first tell what it won't
;;	do. Renumbering these is piece of cake:
;;
;;	    1.1
;;	    1.7.1.5		(1)
;;	    1.5
;;	    1.5.4.1		(2)
;;	    1.5.4.5
;;	    1.9
;;
;;	The result is
;;
;;	    1.1
;;	    1.7.1.5
;;	    1.2
;;	    1.5.4.1
;;	    1.5.4.2
;;	    1.3
;;
;;	tsk-tsk, before you say anything... It went all right.
;;	Now you see what it won't do for you.
;;
;;	.   It can't know that the 1.7.1.5 belongs under previous 1.1,
;;          because no back tracking is done. I won't even try!
;;          [write a separate package if you want that...I won't do it]
;;
;;	.   Same goes here, it can't know that the 1.5.4.1 should actually
;;          start from 1.5.1.1
;;
;;	The thumb rule is, that you _go_ and make sure all the _first_
;;	level headings (those that end to X.X.1.1) are right before doing
;;	renumbering. In the above case, you should have done these before
;;	calling M-x tip-renumber-forward.
;;
;;	.   --> 1.1.1.1
;;	.   --> 1.5.1.1  _AND_ do replace M-% 1.5.4 with 1.5.1
;;
;;	Then all the renumberin would have gone just fine.
;;	Little handy work and this package helps you to number your doc easily.
;;
;;
;;  Renumbering -- be cautious
;;
;;	If you have index section in you file, you have a little problem,
;;	because this package does not know nothing about such things.
;;	If the Index section is at the beginning, just go past it and
;;	use function:
;;
;;	    M-x tip-renumber-forward
;;
;;	Using
;;
;;	    M-x tip-renumber-buffer
;;
;;	Would be disaster. It can only be used for non-index buffers.
;;
;;  Creating index
;;
;;	After you have renumbered all, your old index section is useless,
;;	Just call function
;;
;;	    M-x tip-toc
;;
;;	And copy the showed buffer in place of the old index.
;;
;;  Limitations
;;
;;	Since the numbering is done according to regexp, there is
;;	no way to avoid the following false hit:
;;
;;	    1.1 Overview
;;	    This is highly technical document concerning the latest
;;	    NASA ultrawave reflective shield technique. You should
;;	    refer to chapter:
;;
;;		1.5
;;
;;	    Where the Daddy-Cool portable sondium emission detector is
;;	    described in full...
;;
;;	The Number 1.5 is unfortunately renumbered to 1.2, and possibly
;;	causing headache in the NASA and in the spying countries...
;;	If you know elegant way to prevent these false hits, please
;;	drop me a mail. At this time I haven't much payed attention to this.
;;
;;  Code Note
;;
;;	The renumbering used here uses brute force, so the execution time
;;	is O(n2). If you have more that 30-40 sections, the renumbering
;;	might take 15-40 minutes. If you care to send me more pleasant
;;	numbering I'd be very gratefull and you're name would be carved to
;;	this module. For now, I'm just too lazy to change anything.

;;}}}
;;{{{ history

;; ........................................................ &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; May	29	1997	[jari]		19.28	v1.24		NotReleased
;; - Checkdoc 1.29 clear. defcustom support.
;;
;; May	14	1997	[jari]		19.28	v1.22-1.23	NotReleased
;; - Now uses new minor mode wizard.
;; - Changed all non standard boolean tests to use when and unless
;;
;; Mar	22	1997	[jari]		19.28	v1.21		NotReleased
;; - Clemens Beckstein <beckstei@minet.uni-jena.de> reported sligs bugs
;;   emacsp --> emacs-p. Corrected.
;; - Renamed everything, because the new prefix is more descriptive
;;   tig-*  --> tip-*
;;
;; Mar	22	1997	[jari]		19.28	v1.18-1.20	NotReleased
;; - Added menu support. Changed minor mode installation to use
;;   library functions.
;; - XEmacs overlay support is back.
;;
;; Jan	22	1996	[jari]		19.28	v1.15-1.16	NotReleased
;; - text changes. One function call renamed (Lib change)
;; - Prefix key variable name changed.
;;
;; Oct	1	1996	[jari]		19.28	v1.11-1.13	NotReleased
;; - Rearranged loading. Now uses autoloads as much as possible.
;; - Removed tinymouse.el dependency.
;; - Changed all '(lambda forms to use (function (lambda
;;
;; Aug	21	1996	[jari]		19.28	v1.8-1.10	NotReleased
;; - Corrected all XEmacs 19.14 bytecompiler errors. Made XEmacs compatible.
;; - tip-modeline wasn't interactive as it was supposed to be.
;;
;; Jun	25	1996	[jari]		19.28	v1.7		NotReleased
;; - Renamed functions so that they do not clash with XEmacs 19.14
;;
;; Jun	11	1996	[jari]		19.28	v1.6		NotReleased
;; - Voila! Now it creatres indexes and shows them in a popup etc.
;; - Added sections: 'about making pages', 'renumbering -- be cautious'
;;   'creating index', 'Limitations'.
;; - Added dynamic compilation support. Rewrote the index creating
;;   function. Now the X-popup can be launched from keyboard too.
;; - Rewrote the keybinding installation function, now uses tinymouse.el
;;   if possible for normal page up/down.
;; - Overlay handling made better.
;; - Lots of correctins, new code and fine tuning.
;; - Can now yank to _before_ or _after_ current page. This is handy!
;;
;; Jun	10	1996	[jari]		19.28	v1.4-1.5	NotReleased
;; - Better instrallation instructions. Added autoload notice and
;;   contact info.
;;
;; Jun	10	1996	[jari]		19.28	v1.3		NotReleased
;; - Added very nice renumbering. I like this one a lot!
;;
;; Jun	10	1996	[jari]		19.28	v1.2		NotReleased
;; - Added page length and configurable modeline through user function.
;;
;; Jun	10	1996	[jari]		19.28	v1.1		Created


;;}}}

;;; Code:

;;{{{ setup: require

;;; ......................................................... &require ...

(require 'tinylibm)
(require 'easymenu)

(eval-and-compile
  (if (xemacs-p)
      (or (load "overlay" 'noerr)
	  (message "\n\
tinypage: ** you need XEmacs overlay.el library.
          ** TinyPage may not work correctly without it.")))
  )


(eval-when-compile (use-dynamic-compilation))


(defgroup-tiny TinyPage tip-: tools
  "Minor mode for Handling ^L delimeted text rerions (pages).
  Overview of features
	o   Copy, cut, paste, yank (after/before current page) ^L pages.
	o   Show  page-nbr/page-count/page-size in modeline.
	o   Can renumber numbered header levels, where last level is indicated
	    with number. Eg. \"A.1 A.2\"  or \"1.2.1.1\" \"1.2.1.2\"
	o   Shows popup in X to jump to headings
	o   Create table of contents."
  )

;;}}}
;;{{{ setup: private


;;; .......................................................... &v-mode ...
;;;###autoload (autoload 'tip-mode "tinypage" t t)
;;;
;;;
;;; Prefix keys is "\" by default: this one
;;; was nicely non-shifted and near HP-UX return key.

(defvar tip-:mode-name " ^L"
  "Minor mode name. Changed by program. not user variable.")
(make-variable-buffer-local 'tip-:mode-name)

(eval-and-compile

(ti::m-fmacro-minor-mode-wizard
 "tip-" " ^L" "\\" "Tpage" 'TinyPage "tip-:" ;1-6

  "Paged minor mode. This mode allows you to handle ^L delimited
region as page: you can eg. cut, copy, and select it.

The page counter is _not_ updated all the time in the modeline, because
it'd be too heavy task to monitor user constantly. Please use command
[\\[tip-modeline]] if you want up to date information.

To adjust the user tracking threshold, modify value:

  `tip-:post-command-wakeup-count'

Mode description:
\\{tip-:mode-map}
"

  "TinyPage"
  (progn
    ;; Make sure it's there...
    ;;
    (or (assq 'tip-post-command post-command-hook)
	(add-hook 'post-command-hook 'tip-post-command))

    ;;  - We could leave the hook there because it's no-op if
    ;;    the mode variable is nil.
    ;;  - But i think if user looks at post-command-hook's contents
    ;;    to spot some problems, he appreciates if there is no extra
    ;;    functions in the hook -- only those that need to active
    ;;    in the current buffer/modes.
    ;;
    (if (null tip-mode)
	(remove-hook 'post-command-hook 'tip-post-command)
      ;; Make sure it's there...
      ;;
      (or (assq 'tip-post-command post-command-hook)
	  (add-hook 'post-command-hook 'tip-post-command)))

    (tip-modeline)
    (ti::xe-modeline-update)
    )

   "TinyPage menu"
   (list
    tip-:mode-easymenu-name
     ["Cut"			tip-cut				t]
     ["Copy"			tip-copy			t]
     ["Select"			tip-select			t]
     ["Yank"			tip-yank			t]
     ["Yank before page"        tip-yank-before			t]
     ["Yank after page"		tip-yank-after			t]
     "----"
     ["Renumber buffer"		tip-renumber-buffer		t]
     ["Renumber forward"	tip-renumber-forward		t]
     ["Renumber Level forward"	tip-renumber-level-forward	t]
     "----"
     ["Index"			tip-toc			t]
     ["Index occur"             tip-toc-occur			t]
     ["Index popup"             tip-toc-x-popup		t]
     "----"
     ["Update modeline info"    tip-modeline                    t]
     ["Previous heading"	tip-go-previous			t]
     ["Next heading"            tip-go-next			t]
     ["Scroll down"		scroll-down			t]
     ["Scroll up"		scroll-up			t]
     )

   (progn
     (let* (vec
	    )
       (tinyck-maybe-activate 'emacs)

    (define-key   map p    nil) ;; make prefix available

    (if (eq 1 (length p))
	(define-key   map (concat p p) 'self-insert-command))

    ;; These are the DOS standard keys, mimic them
    ;; Alt c  = copy
    ;; Alt t  = cut
    ;; Alt p  = paste
    ;;
    (define-key   map (concat p "c") 'tip-copy)
    (define-key   map (concat p "t") 'tip-cut)
    (define-key   map (concat p "p") 'tip-yank)
    (define-key   map (concat p "s") 'tip-select)


    ;; ....................................................... yanking ...
    ;;	Emacs users are more familiar with this
    ;;
    (define-key   map (concat p "y")  'tip-yank)


    ;;  Some handy paste commands. Moving pages around
    ;;  See the keyboads: < >       which mean before , after
    ;;  I don't want to use shift...
    ;;
    (define-key   map (concat p ",") 'tip-yank-before)
    (define-key   map (concat p ".") 'tip-yank-after)

    ;; ..................................................... numbering ...
    ;;	key "n" for numbering.
    ;;
    (define-key   map (concat p "nl")  'tip-renumber-level-forward)
    (define-key   map (concat p "nf")  'tip-renumber-forward)
    (define-key   map (concat p "nb")  'tip-renumber-buffer)

    ;; ......................................................... index ...
    ;;	key "i" for indexing
    ;;  I didn't pick "x" for X-popup because it's too far away
    ;;  from the "i" key. The "p" for "popup" is much closer.
    ;;
    (define-key   map (concat p "ii")  'tip-toc)
    (define-key   map (concat p "io")  'tip-toc-occur)
    (define-key   map (concat p "ip")  'tip-toc-x-popup-keyboard)




    ;; ................................................... vector keys ...
    ;; ...................................................... x-events ...
    ;;	Some handy X bindings:
    ;;

    ;; Too bad the delete key is not standard, we have to define many symbols
    ;;
    (setq vec (make-vector-keybinding p 'delete))
    (define-key   map vec  'tip-cut)

    (setq vec (make-vector-keybinding p 'del))
    (define-key   map vec  'tip-cut)

    (setq vec (make-vector-keybinding p 'deletechar))
    (define-key   map vec  'tip-cut)

    (setq vec (make-vector-keybinding p 'hpDeleteChar))	;XEmacs shows this
    (define-key   map vec  'tip-cut)



    (setq vec (make-vector-keybinding p 'backspace))
    (define-key   map vec  'tip-copy)



    (setq vec (make-vector-keybinding p 'insert))
    (define-key   map vec  'tip-yank)

    (setq vec (make-vector-keybinding p 'insertchar))
    (define-key   map vec  'tip-yank)

    (setq vec (make-vector-keybinding p 'hpInsertChar))	;XEmacs 19.14
    (define-key   map vec  'tip-yank)



    ;;	mouse binding "Prefix + i + mouse-1"
    ;;
    (setq vec (make-vector-keybinding p ?i 'mouse-1))
    (define-key   map vec  'tip-toc-x-popup)


    ;; .................................................... go, update ...
    ;;
    (define-key   map (concat p "u") 'tip-modeline)


    (define-key   map [prior]	'tip-go-previous)
    (define-key   map [next]	'tip-go-next)

    (define-key   map [C-prior] 'scroll-down)
    (define-key   map [C-next]  'scroll-up)

    (tinyck-maybe-activate 'emacs 'dis)
    ))
   )
)

;;; ......................................................... &v-hooks ...

(defcustom tip-:load-hook nil
  "*Hook that is run when package is loaded."
  :type 'hook :group 'TinyPage
  )

;;; ....................................................... &v-private ...


(defvar tip-:post-command-wakeup-counter  nil
  "Updated by program.")
(make-variable-buffer-local 'tip-:post-command-wakeup-counter)


(defvar tip-:buffer-toc  "*toc*"
  "Where to create index.")


;;}}}
;;{{{ setup: user

;;; ........................................................ &v-public ...
;;; User configurable

(defcustom tip-:register ?p
  "*Register used for clipboard."
  :type 'character :group 'TinyPage
  )

;;; This is _not_ one char "^L", it is two chars "^" + "L"
;;;
(defcustom tip-:mode-name-string " ^L"
  "*Minor mode name. User variable."
  :type 'string :group 'TinyPage
  )

(defcustom tip-:post-command-wakeup-count  40
  "*How often to wake up to update modeline info.
Don't put too low value, since it slows down Emacs."
  :type 'integer :group 'TinyPage
  )

(defcustom tip-:x-coord  170
  "*Default X menu coordinate."
  :type 'integer :group 'TinyPage
  )

(defcustom tip-:y-coord  170
  "*Default Y menu coordinate."
  :type 'integer :group 'TinyPage
  )

(defcustom tip-:x-popup-line-len 35
  "*Maximum line length in popup."
  :type 'integer :group 'TinyPage
  )


;;  Do not change this ! Unless you know what you do...
;;
(defcustom tip-:renumber-format
  '( "^[ \t]*\\([0-9.]+\\.\\)\\([0-9]+\\)"  1 2)
  "*Regexp for renumbered lines.

Format is

    (REGEXP SAME-LEVEL RENUM-LEVEL),

where the match in RENUM-LEVEL match must return a valid number.
Value SAME-LEVEL is examined only once and the _dots_ that is holds
are counted. The dots tell which mail-level was picked.
Eg, when renumbering the following, only the last number is incremented.

    1.2 Section one is here, this is picked first and examined.
    1.2.1.1 This subsection is skipped, since it's not in the same level.
    1.3 This level will be picked.

There _must_ be dots in the matched level string, because the section
level is calculated by counting the dots.  The following
section numbers won't do:

    1
    2"
  :type '(list
	  (string :tag "Regexp")
	  (integer :tag "Submatch in regexp")
	  (integer :tag "Submatch in regexp")
	  )

  :group 'TinyPage
  )


(defcustom tip-:modeline-function 'tip-update-mode-line
  "*The modeline function that keep it up to date whenever called."
  :type 'integer :group 'TinyPage
  )


;;}}}

;;{{{ version

;;; ....................................................... &v-version ...


(eval-and-compile
(ti::m-fmacro-version-bug-report
 "tinypage.el"
 "tip"
 tip-:version-id
 "$Id: tinypage.el,v 1.24 1997/05/29 18:00:03 jaalto NotReleased $"
 '(tip-:version-id
   map
   tip-:load-hook
   tip-:mode-define-keys-hook
   tip-:mode-hook
   tip-mode
   tip-:mode-name
   tip-:mode-menu
   tip-:buffer-toc
   tip-:register
   tip-:mode-name-string
   tip-:mode-menu-name
   tip-:mode-prefix-key
   tip-:post-command-wakeup-count
   tip-:x-coord
   tip-:y-coord
   tip-:x-popup-line-len
   tip-:modeline-function
   tip-:renumber-format

   tip-:post-command-wakeup-counter
   tip-:version-id
   )))


;;}}}

;;; ########################################################### &Funcs ###

;;{{{ misc, engine funcs

;;; ----------------------------------------------------------------------
;;;
(defun tip-modeline  ()
  "Update modeline information."
  (interactive)
  (funcall tip-:modeline-function))


;;; ----------------------------------------------------------------------
;;;
(defun tip-page-region  (&optional verb)
  "Return region (BEG . END) of page. VERB."
  (interactive)
  (let* (beg
	 end
	 ret
         )
    (save-excursion
      (beginning-of-line)
       (if (looking-at "^[ \t]*\C-l")
	   (setq beg (point))
	 (setq beg (tip-go-next 'back)))

       (when (setq end (tip-go-next))
	 (goto-char end)		;adjust point, do not take ^L
	 (beginning-of-line)
	 (setq end (point))
	 )
       )
    (if (and beg end)
	(setq ret (cons beg end))
      (if verb
	  (message "Couldn't find region")))
    ret
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tip-count-pages ()
  "Count page characters ^L."
  (let* ((count 0)
	 )
    (save-excursion
      (PMIN)
      (while (re-search-forward "^[ \t]*\C-l" nil t)
	(inc count))
      )
    count
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tip-count-lines-in-page ()
  "Count lines."
  (let* ((elt (tip-page-region))
	 (beg (car-safe elt))
	 (end (cdr-safe elt))
	 ret
	 )
    (when elt
      (setq ret (count-lines beg end)))
    ret
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tip-current-page  ()
  "Current page."
  (interactive)
  (let* ((re    "^[ \t]*\C-l")
	 (p     (point))
	 (count 0)
	 )
    (save-excursion
      (PMIN)
      (while (re-search-forward re p t)
	(inc count))
      )
    (if (looking-at re)
	(inc count))
    count
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tip-update-mode-line  ()
  "Update modeline info."
  (interactive)
  (let* ((mode-string  tip-:mode-name-string)
	 pages
	 now
	 lines
	 )
    (setq pages (tip-count-pages))
    (setq now   (tip-current-page))
    (setq lines (tip-count-lines-in-page))
    (setq tip-:mode-name
	  (format  " %s %s/%s/%s" mode-string now pages (or lines "-")))
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tip-post-command  ()
  "Keep page info in modeline up to date"
  (when tip-mode			;only now!
    (if (not (integerp tip-:post-command-wakeup-counter))
	(setq tip-:post-command-wakeup-counter 0))

    (inc tip-:post-command-wakeup-counter)

    (when (eq 0 (% tip-:post-command-wakeup-counter
		   tip-:post-command-wakeup-count))
      (tip-modeline)
      )))


;;; ----------------------------------------------------------------------
;;;
(defun tip-overlay  (act &optional beg end)
  "If ACT is 'hide, hide overlay, otherwise highlight BEG END."
  (let* ((ov
	  (if (boundp 'mouse-drag-overlay) ;Emacs, use this by default
	      'mouse-drag-overlay
	    'primary-selection-extent
	    ))
	 )

    (cond
     ((eq act 'hide)
      (ti::xe-overlay-move ov 1 1)
      )
     (t
      (ti::xe-overlay-move ov beg end)
      (when (emacs-p)
	(push-mark
	 (if (emacs-p)
	     (sfuncall 'overlay-start ov)
	   (sfuncall 'extent-start-position ov))
	 t t)
	(push-mark
	 (if (emacs-p)
	     (sfuncall 'overlay-end ov)
	   (sfuncall 'extent-end-position ov))
	 t t)
	)
      (setq this-command 'set-mark)
      ))
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tip-page-mark-region  (beg end &optional act maybe)
  "Mark region and do some command act.

Input:

  BEG		,region beg
  END		,region end
  ACT		,action name, default is 'copy. Can be also 'cut 'select
  MAYBE		,flag to check is BEG END are valid: if not then do
		nothing. If vallid; then select and do ACT.

References:
  `tip-:register'

Return:
  t		,if successfull
  nil"
  (let* ((doit (if maybe
		   (and beg end)
		 t))
	 (reg tip-:register)
	 )
    (cond
     ((null doit)
      nil)

     (t
      (tip-overlay 'show beg end)
      (cond
       ((eq 'cut act)
	(delete-region beg end)
	(goto-char beg)
	)
       ((memq act '(nil copy))
	(set-register
	 reg
	 (remove-properties (buffer-substring beg end)))
	))
      t					;something done
      ))
    ))

;;}}}
;;{{{ application funcs

;;; .................................................... &applications ...


;;; ----------------------------------------------------------------------
;;;
(defun tip-renumber-level-forward (&optional verb)
  "Renumber current level starting from current line. VERB.
Only the last level number is incremented. Put cursor line above
the level and call this function
    1.2
    *		    <-- cursor here and it'll renumber level 1.2.1.x
    1.2.1.1

References:
  `tip-:renumber-format'"
  (interactive "P")
  (let* ((data   tip-:renumber-format)
	 (re     (nth 0 data))
	 (lev1   (nth 1 data))
	 (lev2   (nth 2 data))

	 nbr
	 counter
	 dots-exact
	 dots
	 level-string
	 orig-level-string
         )
    (VERB)
    (save-excursion
      (while (re-search-forward re nil t)
	(setq level-string (read-match lev1))

	(when level-string
	  (setq dots (count-char-in-string ?. level-string)))

	(when (and (null dots-exact)		;do only once
		   dots)

	  (setq dots-exact		dots
		orig-level-string	level-string)
	  )

;;;	(d! (read-match 0)
;;;	    orig-level-string level-string
;;;	    dots dots-exact
;;;	    counter
;;;	    (read-match lev2)
;;;	    )

	(when (and dots-exact
		   (eq dots-exact dots)	;only same level accepted
		   ;;  Must have same beginning "1.2.1.1" "1.2.1.x"
		   ;;
		   (string= orig-level-string level-string)
		   (setq nbr (read-match lev2))
		   )
	  (if (null counter)		;first value ?
	      (setq counter (string-to-int nbr))
	    (inc counter)
	    ;;  Replace the last number with the right increment
	    (kill-match lev2 (int-to-string counter))
	    ))
	))

    (when verb
      (cond
       ((null counter)
	(message "No matches."))
       (t
	(message "Last heading was %s%s" orig-level-string counter)
	)))
    counter
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tip-renumber-forward  (&optional verb)
  "Renumber all found headings forward. VERB."
  (interactive "P")
  (let* ((data   tip-:renumber-format)
	 (re     (nth 0 data))
	 )
    (VERB)
    ;; Well, we do lot of extra work here, because the
    ;; tip-renumber-level-forward goes alway to the bottom,
    ;; but what the heck... it won't take long.
    ;;
    ;; And code is much cleaner this way.

    (if verb
	(message "Renumbering..."))

    (while (re-search-forward re nil t)
      (beginning-of-line)
      (tip-renumber-level-forward)
      (forward-line 1)
      )
    (if verb
	(message "Renumbering...done"))

    ))


;;; ----------------------------------------------------------------------
;;;
(defun tip-renumber-buffer  ()
  "Renumber all headings in buffer starting from point-min."
  (interactive)
  (save-excursion (PMIN) (tip-renumber-forward 'verb)))



;;; ----------------------------------------------------------------------
;;;
(defsubst tip-get-index-list  ()
  "Return list of strings."
  (let* ((list (ti::b-grep-lines (nth 0 tip-:renumber-format)))
	 elt
	 ret
         )
    (list-loop list elt (push (ti::s-sqz elt) ret))
    ret
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tip-toc (&optional ragged no-show)
  "Create toc to temporary buffer.
Optional argument RAGGED makes the heading to 'hang'.
With Nil RAGGED, the headings are lined up.

NO-SHOW doesn't show buffer after creating table of content.

Return:
  buffer"
  (interactive "P")
  (let* ((list		(ti::b-grep-lines (nth 0 tip-:renumber-format)))
	 (buffer	(use-tmp-buffer tip-:buffer-toc 'clear))
	 elt
	 dots
	 padd
	 heading
	 text
         )
    (job-buffer buffer
      (list-loop list elt
	(setq elt (ti::s-sqz elt))
	(setq heading (strmatget "^[^ \t]+" 0 elt))
	(setq text    (or (strmatget "^[^ \t]+[ \t]+\\(.*\\)" 1 elt)
			  "<no heading found>"))

	;; How to indent this line
	;;
	(setq dots (count-char-in-string ?. heading))
	(if (<= dots 1)
	    (setq padd "")
	  (setq padd (make-string (* 2 (- dots 2)) ?\  )))

	(setq heading (concat padd heading))

	(if ragged
	    (insert heading "    " text "\n")
	  (insert (format "%-16s%s\n" heading text)))

	))

    (unless no-show
      ;; Display it, but do not select/go to it.
      ;;
      (display-buffer buffer)

      (save-excursion-window
	(select-window (get-buffer-window buffer))
	(shrink-window-if-larger-than-buffer))
      )
    buffer
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tip-toc-x-popup-keyboard ()
  "Create index. Show it in X-popup."
  (interactive)
  (tip-toc-x-popup (ti::xe-make-fake-event tip-:x-coord tip-:y-coord)))


;;; ----------------------------------------------------------------------
;;;
(defun tip-toc-x-popup (event)
  "Create index. Show it in X-popup with EVENT."
  (interactive "e")
  (let* ((len    tip-:x-popup-line-len)
	 (title  "Index")
         list
         val
         point
         )
    (cond
     ((null (ti::xe-window-system))
      (message "Sorry, Requires X to use X-popup"))
     (t
      (setq list (tip-get-index-list))
      (setq list (mapcar
		  (function
		   (lambda (x)
		     (str-left x len)))
		  list))

      (when (setq val (ti::xe-popup list event nil title))
	;;  See if we can find the heading...
	(save-excursion-window
	  (PMIN)
	  (if (re-search-forward (regexp-quote val) nil t)
	      (setq point (BOLPP))
	    (message "Cannot find heading..."))
	  ))
      (if point
	  (goto-char point))
      ))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tip-toc-occur ()
  "Create occur buffer for jumpig to Headings easily."
  (interactive)
  (occur (nth 0 tip-:renumber-format)))


;;}}}
;;{{{ interactive funcs

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tip-region-action  (act &optional verb)
  "Execute action ACT. Return t or nil. VERB."
  (let* ((elt	(tip-page-region verb))
	 (beg   (car-safe elt))
	 (end   (cdr-safe elt))
         )
    (VERB)
    (tip-page-mark-region beg end act 'maybe)
    ))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tip-select  (&optional verb)
  "Select page. If sitting on page Marker, use page below. VERB."
  (interactive "P")
  (VERB)
  (and (tip-region-action 'select verb)
       (if verb (message "Page selected.")))
  )

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tip-copy  (&optional verb)
  "Select page. If sitting on page Marker, use page below. VERB."
  (interactive "P")
  (VERB)
  (and (tip-region-action 'copy verb)
       (if verb (message "Page copied.")))
  )


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tip-cut  (&optional verb)
  "Select page. If sitting on page Marker, use page below. VERB."
  (interactive "P")
  (VERB)
  (tip-region-action 'cut verb)
  )

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tip-yank  (&optional verb)
  "Yank page from register. VERB."
  (interactive "P")
  (insert-register tip-:register)
  (tip-overlay 'hide)
  )


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tip-yank-before  (&optional verb)
  "Yank page from register, but _before_ current page. VERB."
  (interactive)
  (VERB)
  (tip-yank-after 'before "Yanked before this page." verb))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tip-yank-after  (&optional before msg verb)
  "Yank page from register, but _after_ current page.
Optionally BEFORE with MSG and VERB."
  (interactive)
  (let* ((msg   (or msg  "Yanked after this page."))
	 )
    (VERB)
    (save-with-marker
      (when (tip-go-next before verb)
	(insert-register tip-:register)
	(tip-overlay 'hide)
	(if (and verb msg)
	    (message msg))
	))
  ))



;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tip-go-previous  (&optional verb)
  "Go to previous page. VERB."
  (interactive)
  (VERB)
  (tip-go-next  'back verb))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tip-go-next  (&optional back verb)
  "Go to next page, optionally BACK. Return point if moved. VERB."
  (interactive)
  (let* ((point (point))
	 func
	 ret
         )
    (VERB)
    (cond
     (back
      (setq func 're-search-backward)
      (beginning-of-line))
     (t
      (setq func 're-search-forward)
      (end-of-line)
      ))

    (unless (setq ret (funcall func "\C-l" nil t))
      (goto-char point))

    (if verb
	(tip-modeline))

    (if (and verb (null ret))
	(message "No more page marks."))
    ret
    ))

;;}}}

(add-hook 'tip-:mode-define-keys-hook 'tip-mode-define-keys)

(provide   'tinypage)
(run-hooks 'tip-:load-hook)

;;; tinypage.el ends here
